@@ -56,6 +56,58 @@
 #include "mkdtemp.h"
 #endif
 
+/* Profiling stuff adapted from gtkfilechooserdefault */
+/* To use run:
+ *  strace -ttt -f -o logfile.txt nautilus-cd-burner
+ */
+
+#undef PROFILE_MAKE_ISO
+#ifdef PROFILE_MAKE_ISO
+
+#define PROFILE_INDENT 4
+static int profile_indent;
+
+static void
+profile_add_indent (int indent)
+{
+        profile_indent += indent;
+        if (profile_indent < 0)
+                g_error ("You screwed up your indentation");
+}
+
+static void
+_ncb_make_iso_profile_log (const char *func,
+                           int         indent,
+                           const char *msg1,
+                           const char *msg2)
+{
+        char *str;
+
+        if (indent < 0)
+                profile_add_indent (indent);
+
+        if (profile_indent == 0)
+                str = g_strdup_printf ("MARK: %s: %s %s %s", G_STRLOC, func, msg1 ? msg1 : "", msg2 ? msg2 : "");
+        else
+                str = g_strdup_printf ("MARK: %s: %*c %s %s %s", G_STRLOC, profile_indent - 1, ' ', func, msg1 ? msg1 : "", msg2 ? msg2 : "");
+
+        access (str, F_OK);
+
+        g_free (str);
+
+        if (indent > 0)
+                profile_add_indent (indent);
+}
+
+#define profile_start(x, y) _ncb_make_iso_profile_log (G_STRFUNC, PROFILE_INDENT, x, y)
+#define profile_end(x, y)   _ncb_make_iso_profile_log (G_STRFUNC, -PROFILE_INDENT, x, y)
+#define profile_msg(x, y)   _ncb_make_iso_profile_log (NULL, 0, x, y)
+#else
+#define profile_start(x, y)
+#define profile_end(x, y)
+#define profile_msg(x, y)
+#endif
+
 GQuark
 nautilus_burn_iso_error_quark (void)
 {
@@ -228,11 +280,12 @@ dir_is_empty (const char *virtual_path)
 	g_free (escaped_path);
 	
 	result = gnome_vfs_directory_open (&handle, uri, GNOME_VFS_FILE_INFO_DEFAULT);
+	g_free (uri);
+
 	if (result != GNOME_VFS_OK) {
-		g_free (uri);
 		return TRUE;
 	}
-	
+
 	info = gnome_vfs_file_info_new ();
 
 	found_file = FALSE;
@@ -447,6 +500,8 @@ create_graft_file (GnomeVFSURI          *uri,
 	GnomeVFSResult           result;
 	gboolean                 stop;
 
+	profile_start ("start", NULL);
+
 	result = gnome_vfs_directory_open_from_uri (&handle,
 						    uri,
 						    GNOME_VFS_FILE_INFO_DEFAULT
@@ -517,6 +572,8 @@ create_graft_file (GnomeVFSURI          *uri,
 
 	gnome_vfs_directory_close (handle);
 	gnome_vfs_file_info_unref (info);
+
+	profile_end ("end", NULL);
 }
 
 static gboolean  
@@ -740,7 +797,7 @@ get_average_rate (GList  **rates,
 	if (g_list_length (*rates) > max_num) {
 		*rates = g_list_delete_link (*rates, *rates);
 	}
-	int_rate = (gint32)round (scale * rate);
+	int_rate = (gint32)ceil (scale * rate);
 
 	*rates = g_list_append (*rates, GINT_TO_POINTER (int_rate));
 
@@ -1151,7 +1208,11 @@ nautilus_burn_make_iso (const char                  *filename,
 	
 	uri = gnome_vfs_uri_new ("burn:///");
 	state.found_file = FALSE;
+
+	profile_start ("start", "creating file list");
 	create_graft_file (uri, NULL, &state);
+	profile_end ("end", "creating file list");
+
 	gnome_vfs_uri_unref (uri);
 
 	fclose (state.graft_file);
@@ -1239,9 +1300,7 @@ nautilus_burn_make_iso (const char                  *filename,
 
 	if (res == -1) {
 		g_warning ("Cannot get free space at %s\n", dirname);
-		g_free (dirname);
 	} else if (iso_size > size) {
-		g_free (dirname);
 		if (flags & NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE) {
 			mkisofs_output.error = g_error_new (NAUTILUS_BURN_ISO_ERROR, 
 							    NAUTILUS_BURN_ISO_ERROR_GENERAL,
@@ -1252,9 +1311,13 @@ nautilus_burn_make_iso (const char                  *filename,
 			mkisofs_output.result = NAUTILUS_BURN_RECORDER_RESULT_RETRY;
 		}
 
+		g_free (dirname);
+
 		goto cleanup;
 	}
 
+	g_free (dirname);
+
 	i = 0;
 	argv [i++] = "mkisofs";
 	argv [i++] = "-r";
@@ -1372,7 +1435,7 @@ nautilus_burn_make_iso (const char                  *filename,
  **/
 int
 nautilus_burn_make_iso_from_drive (const char                  *filename,
-				   const NautilusBurnDrive     *drive,
+				   NautilusBurnDrive           *drive,
 				   NautilusBurnImageCreateFlags flags,
 				   NautilusBurnImageType       *type,
 				   char                       **toc_filename,
