@@ -221,14 +221,6 @@ subscription_get_path (MappingSubscription *sub)
 	return sub->path;
 }
 
-static void *
-subscription_get_userdata (MappingSubscription *sub)
-{
-	g_return_val_if_fail (sub != NULL, NULL);
-
-	return sub->userdata;
-}
-
 static gboolean
 subscription_send_event (MappingSubscription *sub,
 			 const char          *filename,
@@ -995,29 +987,6 @@ connection_remove_subscription (MappingConnection   *connection,
 	subscription_unref (sub);
 }
 
-static GnomeVFSResult
-remove_subscription (MappingConnection   *connection,
-		     MappingSubscription *sub)
-{
-	VirtualRoot *root;
-	VirtualNode *node;
-
-	root = lookup_root (sub->rootname, TRUE);
-
-	node = virtual_node_lookup (root, sub->path, NULL);
-
-	subscription_cancel (sub);
-
-	if (node != NULL) {
-		virtual_node_remove_subscription (node, sub);
-	}
-
-	connection_remove_subscription (connection, sub);
-	subscription_unref (sub);
-
-	return GNOME_VFS_OK;
-}
-
 static void
 connection_free (MappingConnection *connection)
 {
@@ -1029,14 +998,9 @@ connection_free (MappingConnection *connection)
 
 	/* release subscriptions that weren't released by cancelling the monitor */
 	while ((cur = g_list_first (connection->subscriptions)) != NULL) {
-		GnomeVFSResult       res;
 		MappingSubscription *sub = cur->data;
 
-		DEBUG_PRINT ("Removing subscription %p\n", sub);
-		res = remove_subscription (connection, sub);
-		if (GNOME_VFS_OK != res) {
-			DEBUG_PRINT ("Removing connection failed: %d\n", res);
-		}
+		monitor_cancel (sub->rootname, sub->path, connection, NULL);
 	}
 
 	g_io_channel_unref (connection->source);
@@ -1084,8 +1048,7 @@ connection_new (GIOChannel *source)
 
 static MappingSubscription *
 connection_get_subscription (MappingConnection *connection,
-			     const char        *path,
-			     void              *userdata)
+			     const char        *path)
 {
 	GList *l;
 
@@ -1094,10 +1057,8 @@ connection_get_subscription (MappingConnection *connection,
 	for (l = connection->subscriptions; l; l = l->next) {
 		MappingSubscription *sub = l->data;
 
-		if ((strcmp (subscription_get_path (sub), path) == 0)
-		    && (subscription_get_userdata (sub) == userdata)) {
+		if (strcmp (subscription_get_path (sub), path) == 0)
 			return sub;
-		}
 	}
 
 	return NULL;
@@ -1172,18 +1133,29 @@ monitor_cancel (const char        *rootname,
 		MappingConnection *connection,
 		void              *userdata)
 {
+	VirtualRoot         *root;
+	VirtualNode         *node;
 	MappingSubscription *sub;
-	GnomeVFSResult       res;
 
 	DEBUG_PRINT ("Cancelling monitor for %s userdata %p\n", path, userdata);
 
-	sub = connection_get_subscription (connection, path, userdata);
+	root = lookup_root (rootname, TRUE);
 
-	if (!sub) {
+	node = virtual_node_lookup (root, path, NULL);
+
+	if (node == NULL) {
 		return GNOME_VFS_ERROR_NOT_FOUND;
 	}
 
-	res = remove_subscription (connection, sub);
+	sub = connection_get_subscription (connection, path);
+
+	if (!sub)
+		return GNOME_VFS_ERROR_NOT_FOUND;
+
+	subscription_cancel (sub);
+	virtual_node_remove_subscription (node, sub);
+	connection_remove_subscription (connection, sub);
+	subscription_unref (sub);
 
 	return GNOME_VFS_OK;
 }
@@ -1440,7 +1412,7 @@ int
 main (int   argc,
       char *argv[])
 {
-        struct sockaddr_un sin;
+	struct sockaddr_un sin;
 	int                master_socket;
 	GIOChannel        *master_io;
 	GMainLoop         *loop;
@@ -1458,9 +1430,9 @@ main (int   argc,
 #endif
 	
 	if ((master_socket = socket (AF_UNIX, SOCK_STREAM, 0)) == -1) {
-                perror ("Master socket creation failed");
-                exit (1);
-        }
+		perror ("Master socket creation failed");
+		exit (1);
+	}
 	
 	path = mapping_protocol_get_unix_name ();
 
@@ -1470,13 +1442,16 @@ main (int   argc,
 
 	g_unlink (sin.sun_path);
 	if (bind (master_socket, (const struct sockaddr *)&sin, sizeof (sin)) == -1) {
-                perror ("Failed to bind master socket");
-                exit (1);
-        }
+		perror ("Failed to bind master socket");
+		close (master_socket);
+		exit (1);
+	}
+
 	if (listen (master_socket, 5) == -1) {
 		perror ("Failed to listen to master socket");
-                exit (1);
-        }
+		close (master_socket);
+		exit (1);
+	}
 
 	/* Trigger launching app */
 	if (pipe_fd >= 0) {
