@@ -76,6 +76,23 @@ mapping_protocol_data_available (GIOChannel *channel)
 	return (1);
 }
 
+#define BUFFER_LEN 4*1024
+typedef struct
+{
+	GIOChannel *channel;
+	int len;
+	char *ptr;
+	char buffer[BUFFER_LEN];
+} BufferedReadChannel;
+
+static void
+buffered_read_channel_init (BufferedReadChannel * ch, GIOChannel *channel)
+{
+	ch->channel = channel;
+	ch->ptr = NULL;
+	ch->len = 0;
+}
+
 static int
 write_all (GIOChannel *channel,
 	   char	      *buf,
@@ -103,34 +120,49 @@ write_all (GIOChannel *channel,
 }
 
 static int
-read_all (GIOChannel *channel,
-	  char	     *buf,
-	  int         len)
+read_all (BufferedReadChannel *channel,
+	  char	              *buf,
+	  int                  len)
 {
-	int bytes;
 	int bytes_read;
 	int fd;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	fd = g_io_channel_unix_get_fd (channel);
+	while (len) {
+		if (channel->len) {
+			const int tocopy = MIN (len, channel->len);
 
-	bytes = 0;
-	while (bytes < len) {
-		bytes_read = read (fd, buf + bytes, len - bytes);
+			memcpy (buf, channel->ptr, tocopy);
+			channel->len -= tocopy;
+			channel->ptr += tocopy;
+			buf += tocopy;
+			len -= tocopy;
+			
+			if (len == 0)
+				break;
+		}
+
+		/* Still more to read in, so re-fill the buffer */
+		fd = g_io_channel_unix_get_fd (channel->channel);
+
+		bytes_read = read (fd, channel->buffer, BUFFER_LEN);
 		if (bytes_read <= 0) {
 			g_warning ("Read error: bytes %d: %s", bytes_read, strerror (errno));
 			return -1;
 		}
-		bytes += bytes_read;
+
+		/* Reset the counters */
+		channel->ptr = channel->buffer;
+		channel->len = bytes_read;
 	}
 
 	return 0;
 }
 
 static int
-encode_int (GIOChannel *channel,
-	    gint32      val)
+encode_int (GString *str,
+	    gint32   val)
 {
 	char  *ptr;
 	gint32 val2;
@@ -138,12 +170,13 @@ encode_int (GIOChannel *channel,
 	val2 = g_htonl (val);
 
 	ptr = (char *) &val2;
-	return write_all (channel, ptr, 4);
+	g_string_append_len (str, ptr, 4);
+	return 0;
 }
 
 static int
-decode_int (GIOChannel *channel,
-	    gint32     *out_val)
+decode_int (BufferedReadChannel *channel,
+	    gint32              *out_val)
 {
 	unsigned char ptr [4];
 	guint32       val;
@@ -163,27 +196,27 @@ decode_int (GIOChannel *channel,
 }
 
 static int
-encode_string (GIOChannel *channel,
-	       char       *str)
+encode_string (GString *gstring,
+	       char    *str)
 {
 	int len;
 	int res;
 	
 	if (str == NULL) {
-		res = encode_int (channel, -1);
+		res = encode_int (gstring, -1);
 	} else {
 		len = strlen (str);
-		res = encode_int (channel, len);
+		res = encode_int (gstring, len);
 		if (res == 0)
-			res = write_all (channel, str, len);
+			g_string_append (gstring, str);
 	}
 
 	return res;
 }
 
 static int
-decode_string (GIOChannel *channel,
-	       char      **out)
+decode_string (BufferedReadChannel *channel,
+	       char               **out)
 {
 	gint32   len;
 	int      res;
@@ -211,8 +244,8 @@ decode_string (GIOChannel *channel,
 }
 
 static int
-encode_pointer (GIOChannel *channel,
-		gpointer    val)
+encode_pointer (GString *str,
+		gpointer val)
 {
 	int len;
 	int res;
@@ -220,20 +253,20 @@ encode_pointer (GIOChannel *channel,
 	/* We use only local socket so no need for network byte order conversion */
 	
 	if (val == NULL) {
-		res = encode_int (channel, -1);
+		res = encode_int (str, -1);
 	} else {
 		len = sizeof (gpointer);
-		res = encode_int (channel, len);
+		res = encode_int (str, len);
 		if (res == 0) {
-			res = write_all (channel, (char *)&val, len);
+			g_string_append_len (str, (char *)&val, len);
 		}
 	}
 	return res;
 }
 
 static int
-decode_pointer (GIOChannel *channel,
-		gpointer   *out_val)
+decode_pointer (BufferedReadChannel *channel,
+		gpointer            *out_val)
 {
 	int       len;
 	int       res;
@@ -279,26 +312,33 @@ mapping_protocol_request_encode (GIOChannel *channel,
 				 gboolean    option,
 				 void       *userdata)
 {
-	int res;
+	int      res;
+	GString *str;
 
 	g_return_val_if_fail (channel != NULL, -1);
+	
+	str = g_string_new (NULL);
 
-	res = encode_int (channel, operation);
+	res = encode_int (str, operation);
 	if (res != 0) return res;
 	
-	res = encode_string (channel, root);
+	res = encode_string (str, root);
 	if (res != 0) return res;
 
-	res = encode_string (channel, path1);
+	res = encode_string (str, path1);
 	if (res != 0) return res;
 
-	res = encode_string (channel, path2);
+	res = encode_string (str, path2);
 	if (res != 0) return res;
 
-	res = encode_int (channel, option);
+	res = encode_int (str, option);
 	if (res != 0) return res;
 
-	res = encode_pointer (channel, userdata);
+	res = encode_pointer (str, userdata);
+	if (res != 0) return res;
+
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
 	if (res != 0) return res;
 
         g_io_channel_flush (channel, NULL);
@@ -307,31 +347,34 @@ mapping_protocol_request_encode (GIOChannel *channel,
 }
 
 int
-mapping_protocol_request_decode (GIOChannel             *channel,
+mapping_protocol_request_decode (GIOChannel             *iochannel,
 				 MappingProtocolRequest *req)
 {
-	int res;
+	int                 res;
+	BufferedReadChannel channel;
 
-	g_return_val_if_fail (channel != NULL, -1);
+	g_return_val_if_fail (iochannel != NULL, -1);
 	
 	memset (req, 0, sizeof (MappingProtocolRequest));
 	
-	res = decode_int (channel, &req->operation);
+	buffered_read_channel_init (&channel, iochannel);
+
+	res = decode_int (&channel, &req->operation);
 	if (res != 0) return res;
 	
-	res = decode_string (channel, &req->root);
+	res = decode_string (&channel, &req->root);
 	if (res != 0) return res;
 
-	res = decode_string (channel, &req->path1);
+	res = decode_string (&channel, &req->path1);
 	if (res != 0) return res;
 
-	res = decode_string (channel, &req->path2);
+	res = decode_string (&channel, &req->path2);
 	if (res != 0) return res;
 
-	res = decode_int (channel, &req->option);
+	res = decode_int (&channel, &req->option);
 	if (res != 0) return res;
 
-	res = decode_pointer (channel, &req->userdata);
+	res = decode_pointer (&channel, &req->userdata);
 	if (res != 0) return res;
 
 	return 0;
@@ -349,22 +392,20 @@ mapping_protocol_request_destroy (MappingProtocolRequest *req)
    there is a listener out there.
  */
 static int
-encode_handshake (GIOChannel *channel,
-		  char        c)
+encode_handshake (GString *str,
+		  char     c)
 {
-	int res;
+	g_string_append_c (str, c);
 
-	res = write_all (channel, &c, 1);
-
-	return res;
+	return 0;
 }
 
 /* this isn't a real handshake but a more of a sanity check to see if
    there is a listener out there.
  */
 static int
-decode_handshake (GIOChannel *channel,
-		  char        c)
+decode_handshake (BufferedReadChannel *channel,
+		  char                 c)
 {
 	int  res;
 	char buf [2];
@@ -390,29 +431,36 @@ mapping_protocol_reply_encode (GIOChannel           *channel,
 {
 	int res;
 	int i;
+	GString * str;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	res = encode_handshake (channel, 'R');
+	str = g_string_new (NULL);
+
+	res = encode_handshake (str, 'R');
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->result);
+	res = encode_int (str, reply->result);
 	if (res != 0) return res;
 	
-	res = encode_string (channel, reply->path);
+	res = encode_string (str, reply->path);
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->option);
+	res = encode_int (str, reply->option);
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->n_strings);
+	res = encode_int (str, reply->n_strings);
 	if (res != 0) return res;
 
 	for (i = 0; i < reply->n_strings; i++) {
-		res = encode_string (channel, reply->strings [i]);
+		res = encode_string (str, reply->strings [i]);
 		if (res != 0) return res;
 	}
 
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
+	if (res != 0) return res;
+
         g_io_channel_flush (channel, NULL);
 	
 	return 0;
@@ -434,29 +482,32 @@ decode_reply_error (MappingProtocolReply *reply,
 }
 
 int
-mapping_protocol_reply_decode (GIOChannel           *channel,
+mapping_protocol_reply_decode (GIOChannel           *iochannel,
 			       MappingProtocolReply *reply)
 {
 	int i;
 	int res;
+	BufferedReadChannel channel;
 
-	g_return_val_if_fail (channel != NULL, -1);
+	g_return_val_if_fail (iochannel != NULL, -1);
+
+	buffered_read_channel_init (&channel, iochannel);
 
 	memset (reply, 0, sizeof (MappingProtocolReply));
 
-	res = decode_handshake (channel, 'R');
+	res = decode_handshake (&channel, 'R');
 	if (res != 0) return res;
 
-	res = decode_int (channel, &reply->result);
+	res = decode_int (&channel, &reply->result);
 	if (res != 0) return res;
 	
-	res = decode_string (channel, &reply->path);
+	res = decode_string (&channel, &reply->path);
 	if (res != 0) return res;
 
-	res = decode_int (channel, &reply->option);
+	res = decode_int (&channel, &reply->option);
 	if (res != 0) return res;
 
-	res = decode_int (channel, &reply->n_strings);
+	res = decode_int (&channel, &reply->n_strings);
 	if (res != 0) return res;
 
 	if (reply->n_strings > 100000)
@@ -468,7 +519,7 @@ mapping_protocol_reply_decode (GIOChannel           *channel,
 	reply->strings = g_new0 (char *, reply->n_strings);
 	
 	for (i = 0; i < reply->n_strings; i++) {
-		res = decode_string (channel, &reply->strings [i]);
+		res = decode_string (&channel, &reply->strings [i]);
 		if (res != 0) {
 			g_free (reply->strings);
 			reply->strings = NULL;
@@ -503,19 +554,26 @@ mapping_protocol_monitor_event_encode (GIOChannel                  *channel,
 				       MappingProtocolMonitorEvent *event)
 {
 	int res;
+	GString *str;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	res = encode_handshake (channel, 'E');
+	str = g_string_new (NULL);
+
+	res = encode_handshake (str, 'E');
 	if (res != 0) return res;
 
-	res = encode_int (channel, event->type);
+	res = encode_int (str, event->type);
 	if (res != 0) return res;
 
-	res = encode_pointer (channel, event->userdata);
+	res = encode_pointer (str, event->userdata);
 	if (res != 0) return res;
 
-	res = encode_string (channel, event->path);
+	res = encode_string (str, event->path);
+	if (res != 0) return res;
+
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
 	if (res != 0) return res;
 
         g_io_channel_flush (channel, NULL);
@@ -524,25 +582,28 @@ mapping_protocol_monitor_event_encode (GIOChannel                  *channel,
 }
 
 int
-mapping_protocol_monitor_event_decode (GIOChannel                  *channel,
+mapping_protocol_monitor_event_decode (GIOChannel                  *iochannel,
 				       MappingProtocolMonitorEvent *event)
 {
 	int res;
+	BufferedReadChannel channel;
 
-	g_return_val_if_fail (channel != NULL, -1);
+	g_return_val_if_fail (iochannel != NULL, -1);
 
 	memset (event, 0, sizeof (MappingProtocolMonitorEvent));
 
-	res = decode_handshake (channel, 'E');
+	buffered_read_channel_init (&channel, iochannel);
+
+	res = decode_handshake (&channel, 'E');
 	if (res != 0) return res;
 
-	res = decode_int (channel, &event->type);
+	res = decode_int (&channel, &event->type);
 	if (res != 0) return res;
 
-	res = decode_pointer (channel, &event->userdata);
+	res = decode_pointer (&channel, &event->userdata);
 	if (res != 0) return res;
 
-	res = decode_string (channel, &event->path);
+	res = decode_string (&channel, &event->path);
 	if (res != 0) return res;
 
 	return 0;
